import { executeComposer } from "./execute.js";
import * as remoteControl from "./remoteControl.js";

/**
 * @typedef CategoryInfo
 * @property {string} id
 * @property {string} name
 * @property {string} description
 */

/**
 * @typedef {{[K in "styling" | "tools" | string]: CategoryInfo}} CategoryDetails
 */

/**
 * @typedef BooleanDefaultValue
 * @property {"boolean"} type
 * @property {boolean} default
 */

/**
 * @typedef StringDefaultValue
 * @property {"string"} type
 * @property {string} default
 */

/**
 * @typedef NumberDefaultValue
 * @property {"number"} type
 * @property {number} default
 */

/**
 * @typedef BaseQuestion
 * @property {string} question,
 */

/**
 * @typedef {BaseQuestion & BooleanDefaultValue} BooleanQuestion
 * @typedef {BaseQuestion & StringDefaultValue} StringQuestion
 * @typedef {BaseQuestion & NumberDefaultValue} NumberQuestion
 * @typedef {BooleanQuestion | StringQuestion | NumberQuestion} Question
 */

/**
 * @typedef PrettierData
 * @property {boolean} installed
 * @property {unknown} config
 */

/**
 * @typedef TypescriptData
 * @property {boolean} installed
 * @property {unknown} config
 */

/**
 * @typedef SvelteKitData
 * @property {boolean} installed
 * @property {string} libDirectory
 * @property {string} routesDirectory
 */

/**
 * @template {ArgType} Args
 * @typedef Workspace
 * @property {ArgValues<Args>} options
 * @property {string} cwd
 * @property {PrettierData} prettier
 * @property {TypescriptData} typescript
 * @property {SvelteKitData} kit
 */

/**
 * @template {ArgType} Args
 * @typedef {import("@svelte-compose/ast-manipulation").JsAstEditor & Workspace<Args>} ScriptFileEditorArgs
 */

/**
 * @template {ArgType} Args
 * @typedef  ScriptFileType
 * @property {"script"} contentType
 * @property {(editor: ScriptFileEditorArgs<Args>) => void} content
 */

/**
 * @template {ArgType} Args
 * @typedef {ScriptFileType<Args> & BaseFile<Args>} ScriptFile
 */

/**
 * @template {ArgType} Args
 * @typedef {{ content: string } & Workspace<Args>} TextFileEditorArgs
 */

/**
 * @template {ArgType} Args
 * @typedef TextFileType
 * @property {"text"} contentType
 * @property {(editor: TextFileEditorArgs<Args>) => string} content
 */

/**
 * @template {ArgType} Args
 * @typedef {TextFileType<Args> & BaseFile<Args>} TextFile
 */

/**
 * @template {ArgType} Args
 * @typedef {import("@svelte-compose/ast-manipulation").SvelteAstEditor & Workspace<Args>} SvelteFileEditorArgs
 */

/**
 * @template {ArgType} Args
 * @typedef SvelteFileType
 * @property {"svelte"} contentType
 * @property {(editor: SvelteFileEditorArgs<Args>) => void} content
 */

/**
 * @template {ArgType} Args
 * @typedef {SvelteFileType<Args> & BaseFile<Args>} SvelteFile
 */

/**
 * @template {ArgType} Args
 * @typedef {{ data: any} & Workspace<Args>} JsonFileEditorArgs
 */

/**
 * @template {ArgType} Args
 * @typedef JsonFileType
 * @property {"json"} contentType
 * @property {(editor: JsonFileEditorArgs<Args>) => void} content
 */

/**
 * @template {ArgType} Args
 * @typedef {JsonFileType<Args> & BaseFile<Args>} JsonFile
 */

/**
 * @template {ArgType} Args
 * @typedef {import("@svelte-compose/ast-manipulation").HtmlAstEditor & Workspace<Args>} HtmlFileEditorArgs
 */

/**
 * @template {ArgType} Args
 * @typedef  HtmlFileType
 * @property {"html"} contentType
 * @property {(editor: HtmlFileEditorArgs<Args>) => void} content
 */

/**
 * @template {ArgType} Args
 * @typedef {HtmlFileType<Args> & BaseFile<Args>} HtmlFile
 */

/**
 * @template {ArgType} Args
 * @typedef {import("@svelte-compose/ast-manipulation").CssAstEditor & Workspace<Args>} CssFileEditorArgs
 */

/**
 * @template {ArgType} Args
 * @typedef  CssFileType
 * @property {"css"} contentType
 * @property {(editor: CssFileEditorArgs<Args>) => void} content
 */

/**
 * @template {ArgType} Args
 * @typedef {CssFileType<Args> & BaseFile<Args>} CssFile
 */

/**
 * @template {ArgType} Args
 * @typedef BaseFile
 * @property {(options: Workspace<Args>) => string} name
 * @property {ConditionDefinition<Args>} [condition]
 */

/**
 * @typedef {Record<string, Question>} ArgType
 */

/**
 * @template {ArgType} Args
 * @typedef  {{
 *   [K in keyof Args]: Args[K]["type"] extends "string" ? string: Args[K]["type"] extends "boolean"? boolean: Args[K]["type"] extends "number"? number: never
 * }} ArgValues
 */

/**
 * @template {ArgType} Args
 * @typedef PackageDefinition
 * @property {string} name
 * @property {string} version
 * @property {boolean} dev
 * @property {ConditionDefinition<Args>} [condition]
 */

/**
 * @template {ArgType} Args
 * @typedef {(workspace: Workspace<Args>) => boolean} ConditionDefinition
 */

/**
 * @template {ArgType} Args
 * @typedef {ScriptFile<Args>
 *  | TextFile<Args>
 *  | SvelteFile<Args>
 *  | JsonFile<Args>
 *  | HtmlFile<Args>
 *  | CssFile<Args>
 * } FileTypes
 */

/**
 * @typedef WebsiteMetadata
 * @property {string} website.logo
 * @property {string[]} website.keywords
 * @property {string} documentation
 */

/**
 * @typedef ComposerConfigEnvironments
 * @property {boolean} svelte
 * @property {boolean} kit
 */

/**
 * @typedef ComposerConfigMetadata
 * @property {string} id
 * @property {string} package
 * @property {string} version
 * @property {string} name
 * @property {string} description
 * @property {CategoryInfo} category
 * @property {ComposerConfigEnvironments} environments
 * @property {WebsiteMetadata} [website]
 */

/**
 * @template {ArgType} Args
 * @typedef ComposerConfig
 * @property {ComposerConfigMetadata} metadata
 * @property {Args?} options
 * @property {PackageDefinition<Args>[]} packages
 * @property {FileTypes<Args>[]} files
 * @property {(options: Workspace<Args>) => Promise<void>} [installHook]
 * @property {(options: Workspace<Args>) => Promise<void>} [uninstallHook]
 */

/**
 * @template {ArgType} Args
 * @param {ComposerConfig<Args>} config
 * @returns {ComposerConfig<Args>}
 */
export function defineComposerConfig(config) {
    return config;
}

/**
 * @template {ArgType} Args
 * @typedef Composer
 * @property {ComposerConfig<Args>} config
 * @property {ComposerCheckConfig<Args>} checks
 * @property {ComposerTestConfig<Args>} [tests]
 */

/**
 * @typedef {Composer<Record<string, Question>>} ComposerWithoutExplicitArgs
 */

/**
 * @template {ArgType} Args
 * @param {ComposerConfig<Args>} config
 * @param {ComposerCheckConfig<Args>} checks
 * @param {ComposerTestConfig<Args>} [tests]
 * @returns {Composer<Args>}
 */
export function defineComposer(config, checks, tests) {
    const remoteControlled = remoteControl.isRemoteControlled();
    if (!remoteControlled) {
        executeComposer(config, checks);
    }

    return { config, checks, tests };
}

/**
 * @typedef Tests
 * @property {(selector: string, property: string, expectedValue: string) => Promise<void>} expectProperty
 * @property {(selector: string) => Promise<void>} elementExists
 * @property {(selector: string, waitForNavigation: boolean) => Promise<void>} click
 * @property {(path: string) => Promise<void>} expectUrlPath
 */

/**
 * @template {ArgType} Args
 * @typedef TestDefinition
 * @property {string} name
 * @property {(tests: Tests) => Promise<void>} run
 * @property {(options: ArgValues<Args>) => boolean} [condition]
 */

/**
 * @template {ArgType} Args
 * @typedef {{
 *   [K in keyof Args]: boolean | number | string
 * }} OptionValues
 */

/**
 * @template {ArgType} Args
 * @typedef ComposerTestConfig
 * @property {FileTypes<Args>[]} files
 * @property {TestDefinition<Args>[]} tests
 * @property {Args} options
 * @property {ArgValues<Args>[]} optionValues
 */

/**
 * @template {ArgType} Args
 * @param {ComposerTestConfig<Args>} tests
 * @returns {ComposerTestConfig<Args>}
 */
export function defineComposerTests(tests) {
    return tests;
}

/**
 * @template {ArgType} Args
 * @param {Args} options
 * @returns {Args}
 */
export function defineComposerOptions(options) {
    return options;
}

/**
 * @typedef PreInstallationCheck
 * @property {string} name
 * @property {(workingDirectory: string) => boolean} run
 */

/**
 * @template {ArgType} Args
 * @typedef PostInstallationCheck
 * @property {string} name
 * @property {(workspace: Workspace<Args>) => boolean} run
 */

/**
 * @template {ArgType} Args
 * @typedef ComposerCheckConfig
 * @property {PreInstallationCheck[]} [preInstallation]
 * @property {PostInstallationCheck<Args>[]} postInstallation
 * @property {Args} options
 */

/**
 * @template {ArgType} Args
 * @param {ComposerCheckConfig<Args>} checks
 * @returns {ComposerCheckConfig<Args>}
 */
export function defineComposerChecks(checks) {
    return checks;
}
